/*
 *  sk.c Ver 0.3 : for sokoban.el
 *
 *
 *   Created by Shigeki.Morimoto (mori@freedom.co.jp)
 *
 *    These codes are copied and modified from play.c in xsokoban-3.1
 *
 *        Idea taken from:    xsokoban-3.1 <andru@lcs.mit.edu>
 *
 *			Last Modified:    Sep 7, 1994
 *
 *  Ver 0.3 : Loop.
 *  Ver 0.2 : If (= start end) then man move.
 *  Ver 0.1 : Just make.
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>

#define 	MAX_X					20
#define 	MAX_Y					20
#define		BADMOVE					(MAX_X * MAX_Y)
#define		UNVISITED				(BADMOVE + 1)

#define		START_CMD				"S:"
#define		END_CMD					"E:"

typedef struct{
	int		x;
	int		y;
	int		d;
}LIST;

/*                  Up Down Left Right */
static int Dx[] = { 0, 0,  -1,   1};
static int Dy[] = {-1, 1,   0,   0};
static char *moveFunctions[] = {
	"(sokoban-move-up)",
	"(sokoban-move-down)",
	"(sokoban-move-left)",
	"(sokoban-move-right)",
};

#define 	WALL_SQUARE				'#'						/* Pattern for wall square */
#define 	GOLD_SQUARE				'>'						/* Pattern for gold packet square */
#define 	GOAL_SQUARE				'.'						/* Pattern for goal square */
#define 	PLAYER_SQUARE			'['						/* Pattern for player square */
#define		GOLD_ON_GOAL_SQUARE		')'						/* Pattern for gold on goal square */
#define		PLAYER_ON_GOAL_SQUARE	'{'						/* Pattern for player on goal square */
#define 	EMPTY_SQUARE			' '						/* Pattern for an empty square */

#define 	ISCLEAR(x,y)			\
	(EMPTY_SQUARE  == map[x][y] || GOAL_SQUARE == map[x][y])
#define		ISPLAYER(x,y)			\
	(PLAYER_SQUARE == map[x][y] || PLAYER_ON_GOAL_SQUARE == map[x][y])
#define		ISPACKET(x,y)			\
	(GOLD_SQUARE   == map[x][y] || GOLD_ON_GOAL_SQUARE == map[x][y])
#define		SETPACKET(x,y)			\
	(map[x][y] = ((EMPTY_SQUARE == map[x][y]) ? GOLD_SQUARE  : GOLD_ON_GOAL_SQUARE))
#define		CLEARPACKET(x,y)		\
	(map[x][y] = ((GOLD_SQUARE  == map[x][y]) ? EMPTY_SQUARE : GOAL_SQUARE))
#define		SETPLAYER(x,y)			\
	(map[x][y] = ((EMPTY_SQUARE == map[x][y]) ? PLAYER_SQUARE  : PLAYER_ON_GOAL_SQUARE))
#define		CLEARPLAYER(x,y)		\
	(map[x][y] = ((PLAYER_SQUARE == map[x][y]) ? EMPTY_SQUARE  : GOAL_SQUARE))


#define		CAN_GO(x,y,d)			\
    (ISCLEAR(x + Dx[d], y + Dy[d]) &&		\
	 ISCLEAR(x - Dx[d], y - Dy[d]) &&		\
	 PossibleToReach(x - Dx[d], y - Dy[d]))


int x0, y0, x1, y1;

char map[MAX_X][MAX_Y];
int dmap[MAX_X][MAX_Y][4];
int pmap[MAX_X][MAX_Y][4];
int findmap[MAX_X][MAX_Y];

int player_X, player_Y;

extern void main();

static int mapRead();
static void slove();
static int	PushFromDir();
static void	RunTo();
static int	PossibleToReach();
static void	FindTarget();
static void	MakeMove();
static void	PrepareDistanceMap();
static void FillTraceMap();

void main (argc, argv)
	int argc;
	char **argv;
{
	char buf[BUFSIZ];

	for(;;){
		while(NULL != fgets(buf, sizeof(buf), stdin)){
			char dummy[BUFSIZ];
			
			if(0 == strncmp(buf, START_CMD, sizeof(START_CMD)-1) &&
			   5 == sscanf(buf, "%s %d %d %d %d", dummy, &x0, &y0, &x1, &y1)){
				if(0 == mapRead()){
					slove();
				}
			}else{
				printf("(error \"Bad arguments\")\n");
			}

			printf(END_CMD "\n");
			fflush(stdout);
		}
		clearerr(stdin);
		usleep(100*1000);
	}
	/* NOTREACHED */
}

static int mapRead()
{
	int x, y;
	char buf[BUFSIZ];

	y = 0;
	for(;;){
		for(    ;NULL != fgets(buf, sizeof(buf), stdin);y ++){
			if(0 == strncmp(buf, END_CMD, sizeof(END_CMD)-1)){
				return 0;
			}
			
			for(x = 0;x < MAX_X;x ++){
				map[x][y] = EMPTY_SQUARE;
			}
			
			for(x = 0;'\0' != buf[x] && '\0' != buf[x+1];x +=2){
				map[x/2][y] = buf[x];
				if(PLAYER_SQUARE == buf[x] ||
				   PLAYER_ON_GOAL_SQUARE == buf[x]){
					player_X = x/2;
					player_Y = y;
				}
			}
		}
		clearerr(stdin);
		usleep(100*1000);
	}
}

static void slove()
{
	int result;
	int k, r, d, x, y;
	
	if(x0 == x1 && y0 == y1){
		/* Only man move. */
		
		if(ISCLEAR(x0, y0) &&
		   PossibleToReach(x0, y0)){
			printf("(progn\n");
			RunTo(x0, y0);
			printf(")\n");
		}else{
			printf("(error \"BAD MOVE 0 x0=%d y0=%d map=[%c]\")\n", x0, y0, map[x0][y0]);
		}
		return;
	}
	
	if(!ISCLEAR(x1, y1) && !ISPLAYER(x1, y1)){
		printf("(error \"BAD MOVE 1 x1=%d y1=%d\")\n", x1, y1);
		return;
	}
	
	CLEARPLAYER(player_X, player_Y);
	
	PrepareDistanceMap();
	
	/* Try to go from the start position in all four directionsn */
	result = BADMOVE;
	for(k = 0;k < 4;k ++){
		if(CAN_GO(x0, y0, k)){
			r = PushFromDir(x0, y0, k, 1);
			if(r < result) {
				d = k;
				result = r;
			}
		}
	}
	
	/* Put the player on the board again */
	SETPLAYER(player_X, player_Y);
	
	
	/* If we found a way (i.e. result < BADMOVE) then follow the route */
	if(result < BADMOVE){
		printf("(progn\n");
		
		x = x0;
		y = y0;
		while(x != x1 || y != y1){
			d = dmap[x][y][d];
			RunTo(x - Dx[d], y - Dy[d]);
			x += Dx[d];
			y += Dy[d];
			
			MakeMove(d);
		}
		printf(")\n");
	}else{
		/* Otherwise */
		printf("(error \"BAD MOVE 2\")\n");
	}
	
	return;
}


static int PushFromDir (x, y, direction, dist)
	int x;
	int y;
	int direction;
	int dist;
{
    int r, res, d, fx, fy;
	
    fx = player_X;
    fy = player_Y;
	
	/* Have we been here before? */
    if(UNVISITED != pmap[x][y][direction]){
		if(BADMOVE == pmap[x][y][direction]){
			return BADMOVE;
		}else{
			return dist + pmap[x][y][direction] - 1;
		}
    }
	
    /* Have we found the target? */
    if(x == x1 && y == y1) {
		pmap[x][y][direction] = 1;
		return dist;
    }
	
	/* Try going in all four directions. Choose the shortest path, if any */
    res = BADMOVE;
    pmap[x][y][direction] = BADMOVE;
	
    for(d = 0;d < 4;d ++){
		player_X = fx;
		player_Y = fy;
		
		if(CAN_GO(x, y, d)){
			CLEARPACKET(x, y);
			SETPACKET(x + Dx[d], y + Dy[d]);
			player_X = x;
			player_Y = y;
			
			r = PushFromDir(x + Dx[d], y + Dy[d], d, dist + 1);
			if(r < res){
				dmap[x][y][direction] = d;
				res = r;
			}
			SETPACKET(x, y);
			CLEARPACKET(x + Dx[d], y + Dy[d]);
		}
    }
    pmap[x][y][direction] = (res == BADMOVE) ? BADMOVE : res - dist + 1;
    player_X = fx;
    player_Y = fy;
	
    return res;
}

static void RunTo (x, y)
	int x;
	int y;
{
    int cx, cy, k, nx, ny, next;
	
    FillTraceMap();
	
	/* flood fill search to find a shortest path to the push point. */
    FindTarget(x, y, 0);
	
	/* if we didn't make it back to the players position, there is no valid path to that place. */
    if(BADMOVE != findmap[player_X][player_Y]){
		/* we made it back, so let's walk the path we just built up  */
		cx = player_X;
		cy = player_Y;
		
		while(findmap[cx][cy]){
			next = findmap[cx][cy] - 1;
			for(k = 0;k < 4;k ++){
				nx = cx + Dx[k];
				ny = cy + Dy[k];
				
				if(findmap[nx][ny] == next){
					MakeMove(k);
					cx = nx;
					cy = ny;
					break;
				}
			}
		}
	}
}


static int PossibleToReach (x, y)
	int x;
	int y;
{
    FillTraceMap();
	
    FindTarget(x, y, 0);
	
    return (findmap[player_X][player_Y] != BADMOVE);
}

static void FindTarget (px, py, pathlen)
	int px;
	int py;
	int pathlen;
{
    int 	i, j, nx, ny;
	int 	num_next;
	LIST	*next;
	int		num_NEXT;
	LIST	*NEXT;
	static LIST	one[BADMOVE], two[BADMOVE];
	
    pathlen = 0;
	
    findmap[px][py] = pathlen;
	
	next = one;
	num_next = 0;
	NEXT = two;
	num_NEXT = 0;
	
	next[num_next].x = px;
	next[num_next].y = py;
	num_next ++;
	
    pathlen ++;
	
    for(;;){
		num_NEXT = 0;
		for(i = 0;i < num_next;i ++){
			px = next[i].x;
			py = next[i].y;
			
			for(j = 0;j < 4;j ++){
				nx = px + Dx[j];
				ny = py + Dy[j];
				
				if((ISCLEAR(nx,ny) || ISPLAYER(nx,ny)) &&
				   pathlen < findmap[nx][ny]){
					findmap[nx][ny] = pathlen;
					
					NEXT[num_NEXT].x = nx;
					NEXT[num_NEXT].y = ny;
					num_NEXT ++;
				}
			}
		}
		if(0 == num_NEXT){
			break;
		}
		
		{
			LIST *tmp;
			
			num_next = num_NEXT;
			tmp = next;
			next = NEXT;
			NEXT = tmp;
			
		}
		pathlen ++;
    }
}

static void MakeMove (d)
	int d;
{
	int nx, ny;
	int mx, my;
	
	nx = player_X + Dx[d];
	ny = player_Y + Dy[d];
	mx = nx       + Dx[d];
	my = ny       + Dy[d];
	
	if(ISPACKET(nx,ny)){
		CLEARPACKET(nx,ny);
		SETPACKET(mx,my);
		printf(";; Push\n");
	}
	CLEARPLAYER(player_X, player_Y);
	SETPLAYER(nx, ny);
	
	player_X = nx;
	player_Y = ny;
	
	printf("(sit-for 0)\n");
    printf("%s\n", moveFunctions[d]);
}

static void PrepareDistanceMap()
{
	int x, y, k;
	
    for(x = 0;x < MAX_X;x ++){
		for(y = 0;y < MAX_Y;y ++){
			int p;
			
			if(ISCLEAR(x,y) || (x == x0 && y == y0)){
				p = UNVISITED;
			}else{
				p = BADMOVE;
			}
			for(k = 0;k < 4;k ++){
				pmap[x][y][k] = p;
			}
		}
    }
}

static void FillTraceMap()
{
    int x, y;
	
	for(x = 0;x < MAX_X;x ++){
		for(y = 0;y < MAX_Y;y ++){
			findmap[x][y] = BADMOVE;
		}
	}
}
